package io.github.palexdev.validationdemo.validation;

import java.util.*;
import java.util.function.Predicate;

public class Validator<T> implements BaseValidator<T> {
	private final Map<String, Constraint<T>> constraints = new HashMap<>();
	private final List<BaseValidator<T>> dependencies = new ArrayList<>();
	private boolean failFast = false;
	private boolean sortBySeverity = false;

	public Validator<T> addConstraint(String name, Constraint<T> constraint) {
		constraints.put(name, constraint);
		return this;
	}

	public Validator<T> addConstraint(String name, Severity severity, String message, Predicate<T> condition) {
		constraints.put(name, Constraint.of(severity, message, condition));
		return this;
	}

	public Validator<T> removeConstraint(String name) {
		constraints.remove(name);
		return this;
	}

	@SafeVarargs
	public final Validator<T> addDependencies(BaseValidator<T>... validators) {
		Collections.addAll(dependencies, validators);
		return this;
	}

	@SafeVarargs
	public final Validator<T> removeDependencies(BaseValidator<T>... validators) {
		dependencies.removeAll(List.of(validators));
		return this;
	}

	@Override
	public List<Constraint<T>> validate(T value) {
		List<Constraint<T>> invalidConstraints = new ArrayList<>();

		for (BaseValidator<T> dependency : dependencies) {
			invalidConstraints.addAll(dependency.validate(value));
			if (dependency.doesFailFast()) return invalidConstraints;
		}

		for (Map.Entry<String, Constraint<T>> entry : constraints.entrySet()) {
			Constraint<T> constraint = entry.getValue();
			if (!constraint.checkOn(value)) {
				invalidConstraints.add(constraint);
				if (failFast) return invalidConstraints;
			}
		}
		if (sortBySeverity) invalidConstraints.sort(Comparator.comparing(Constraint::getSeverity));
		return invalidConstraints;
	}

	@Override
	public boolean doesFailFast() {
		return failFast;
	}

	@Override
	public Validator<T> setFailFast(boolean failFast) {
		this.failFast = failFast;
		return this;
	}

	@Override
	public boolean sortBySeverity() {
		return sortBySeverity;
	}

	@Override
	public Validator<T> setSortBySeverity(boolean sortBySeverity) {
		this.sortBySeverity = sortBySeverity;
		return this;
	}
}
