package io.github.palexdev.validationdemo.validation.automatic;

import io.github.palexdev.validationdemo.others.ValidProperty;
import io.github.palexdev.validationdemo.others.When;
import io.github.palexdev.validationdemo.validation.Severity;
import javafx.beans.binding.BooleanExpression;
import javafx.beans.property.ReadOnlyBooleanProperty;

import java.util.*;

public class AutomaticValidator implements BaseAutomaticValidator {
	private final Map<String, AutomaticConstraint> constraints = new HashMap<>();
	private final List<BaseAutomaticValidator> dependencies = new ArrayList<>();
	private final ValidProperty valid = new ValidProperty();
	private Runnable onInvalid;
	private boolean sortBySeverity = false;

	public AutomaticValidator() {
		When.onInvalidated(valid).then(value -> onInvalid()).listen();
	}

	public AutomaticValidator addConstraint(String name, AutomaticConstraint constraint) {
		constraints.put(name, constraint);
		return this;
	}

	public AutomaticValidator addConstraint(String name, Severity severity, String message, BooleanExpression condition) {
		AutomaticConstraint constraint = AutomaticConstraint.of(severity, message, condition);
		return addConstraint(name, constraint);
	}

	public AutomaticValidator removeConstraint(String name) {
		AutomaticConstraint remove = constraints.remove(name);
		remove.stop();
		return this;
	}

	public AutomaticValidator addDependencies(BaseAutomaticValidator... validators) {
		Collections.addAll(dependencies, validators);
		return this;
	}

	public AutomaticValidator removeDependencies(BaseAutomaticValidator... validators) {
		dependencies.removeAll(List.of(validators));
		return this;
	}

	public void onInvalid() {
		if (onInvalid != null) onInvalid.run();
	}

	public AutomaticValidator setOnInvalid(Runnable onInvalid) {
		this.onInvalid = onInvalid;
		return this;
	}

	public void update() {
		boolean forceFire;

		for (BaseAutomaticValidator dependency : dependencies) {
			if (!dependency.isValid()) {
				forceFire = !isValid();
				setValid(false);
				if (forceFire) valid.fireValueChangedEvent();
				return;
			}
		}

		for (AutomaticConstraint constraint : constraints.values()) {
			if (!constraint.isValid()) {
				forceFire = !isValid();
				setValid(false);
				if (forceFire) valid.fireValueChangedEvent();
				return;
			}
		}

		forceFire = isValid();
		setValid(true);
		if (forceFire) valid.fireValueChangedEvent();
	}

	@Override
	public List<AutomaticConstraint> validate() {
		List<AutomaticConstraint> invalidConstraints = new ArrayList<>();

		for (BaseAutomaticValidator dependency : dependencies) {
			if (!dependency.isValid()) invalidConstraints.addAll(dependency.validate());
		}

		for (Map.Entry<String, AutomaticConstraint> entry : constraints.entrySet()) {
			AutomaticConstraint constraint = entry.getValue();
			if (!constraint.isValid()) invalidConstraints.add(constraint);
		}

		if (sortBySeverity()) invalidConstraints.sort(Comparator.comparing(AutomaticConstraint::getSeverity));
		return invalidConstraints;
	}

	@Override
	public boolean sortBySeverity() {
		return sortBySeverity;
	}

	@Override
	public BaseAutomaticValidator setSortBySeverity(boolean sortBySeverity) {
		this.sortBySeverity = sortBySeverity;
		return this;
	}

	public boolean isValid() {
		return valid.get();
	}

	public ReadOnlyBooleanProperty validProperty() {
		return valid.getReadOnlyProperty();
	}

	protected void setValid(boolean valid) {
		this.valid.set(valid);
	}
}
