package io.github.palexdev.validationdemo.validation2;

import io.github.palexdev.validationdemo.others.When;
import io.github.palexdev.validationdemo.validation.Severity;
import javafx.beans.InvalidationListener;
import javafx.beans.binding.BooleanExpression;
import javafx.beans.property.ReadOnlyBooleanProperty;
import javafx.beans.property.ReadOnlyBooleanWrapper;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.function.BiConsumer;

public class Validator {
	private final ObservableList<Constraint> constraints = FXCollections.observableArrayList();
	private final ObservableList<Validator> dependencies = FXCollections.observableArrayList();
	private final ReadOnlyBooleanWrapper valid = new ReadOnlyBooleanWrapper();
	private BiConsumer<Boolean, List<Constraint>> onUpdated;
	private boolean sortBySeverity = false;

	public Validator() {
		constraints.addListener((InvalidationListener) invalidated -> update());
	}

	public Validator constraint(Constraint constraint) {
		When.onInvalidated(constraint.getCondition()).then(value -> update()).listen();
		constraints.add(constraint);
		return this;
	}

	public Validator constraint(Severity severity, String message, BooleanExpression condition) {
		return constraint(Constraint.of(severity, message, condition));
	}

	public Validator removeConstraint(Constraint constraint) {
		if (constraints.remove(constraint)) {
			When.disposeFor(constraint.getCondition());
		}
		return this;
	}

	public List<Constraint> validate() {
		List<Constraint> invalidConstraints = new ArrayList<>();
		for (Validator dependency : dependencies) if (!dependency.isValid()) invalidConstraints.addAll(dependency.validate());
		for (Constraint constraint : constraints) if (!constraint.isValid()) invalidConstraints.add(constraint);
		if (sortBySeverity) invalidConstraints.sort(Comparator.comparing(Constraint::getSeverity));
		return invalidConstraints;
	}

	public void update() {
		boolean valid = true;
		for (Validator dependency : dependencies) {
			valid = valid && dependency.isValid();
		}
		for (Constraint constraint : constraints) {
			valid = valid && constraint.isValid();
		}
		setValid(valid);
		onUpdated();
	}

	protected void onUpdated() {
		if (onUpdated != null) {
			onUpdated.accept(isValid(), validate());
		}
	}

	public boolean isValid() {
		return valid.get();
	}

	public ReadOnlyBooleanProperty validProperty() {
		return valid.getReadOnlyProperty();
	}

	protected void setValid(boolean valid) {
		this.valid.set(valid);
	}

	public BiConsumer<Boolean, List<Constraint>> getOnUpdated() {
		return onUpdated;
	}

	public Validator setOnUpdated(BiConsumer<Boolean, List<Constraint>> onUpdated) {
		this.onUpdated = onUpdated;
		return this;
	}

	public boolean isSortBySeverity() {
		return sortBySeverity;
	}

	public Validator setSortBySeverity(boolean sortBySeverity) {
		this.sortBySeverity = sortBySeverity;
		return this;
	}
}
